#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "libsox_stub.h"
#include "libsrc_stub.h"

#define CONVERTER SRC_SINC_MEDIUM_QUALITY
//#define CONVERTER SRC_SINC_BEST_QUALITY

//////////////////////////////////////////////////////////////////////////////
typedef struct map {
  char *name;
  int converter;
} map_t;

static map_t map[]={
  { "fast",SRC_SINC_FASTEST },
  { "medium",SRC_SINC_MEDIUM_QUALITY },
  { "best",SRC_SINC_BEST_QUALITY },
  { NULL }
};

static int map2con(map_t *map, char *name)
{
  while (NULL!=map->name) {
    if (0==strcmp(map->name,name))
	  return map->converter;

	++map;
  }

  return CONVERTER;
}

//////////////////////////////////////////////////////////////////////////////
typedef struct {
  size_t size;
  float *data;
} buf_t;

static buf_t *buf_init(buf_t *buf)
{
  buf->size=0;
  buf->data=NULL;

  return buf;
}

static buf_t *buf_cleanup(buf_t *buf)
{
  if (NULL!=buf->data)
    free(buf->data);

  return buf;
}

static buf_t *buf_alloc(buf_t *buf, size_t size)
{
  if (0==buf->size) {
    if (NULL==(buf->data=malloc(size*sizeof *buf->data)))
      goto error;

    buf->size=size;
  }
  else if (buf->size<size) {
    int n=2;
    float *data;

    while (n*buf->size<size)
      n*=2;

    size=n*buf->size;

    if (NULL==(data=realloc(buf->data,size*sizeof *buf->data)))
      goto error;

    buf->size=size;
    buf->data=data;
  }

  return buf;
error:
  return NULL;
}

//////////////////////////////////////////////////////////////////////////////
typedef struct priv {
  double rate;
  int converter;
  buf_t input;
  buf_t output;
  SRC_STATE *src_state;
  SRC_DATA src_data;
} priv_t;

static int getopts(sox_effect_t *effp, int argc, char *argv[])
{
  priv_t *p=effp->priv;
  int n=1;

  if (argc<2)
    return SOX_EOF;

  p->rate=atof(argv[n++]);
  p->converter=n<argc?map2con(map,argv[n++]):CONVERTER;

  buf_init(&p->input);
  buf_init(&p->output);

  return SOX_SUCCESS;
}

static int start(sox_effect_t *effp)
{
  priv_t *p=effp->priv;
  int channels=effp->in_signal.channels;

  if (NULL==(p->src_state=__src_new(p->converter,channels,NULL)))
    return SOX_EOF;

  return SOX_SUCCESS;
}

static int flow(sox_effect_t *effp, sox_sample_t const *ibuf,
    sox_sample_t *obuf, size_t *isamp, size_t *osamp)
{
  int channels=effp->in_signal.channels;
  double rate=effp->in_signal.rate;
  priv_t *p=effp->priv;

  if (NULL==buf_alloc(&p->input,*isamp))
    return SOX_EOF;

  if (NULL==buf_alloc(&p->output,*osamp))
    return SOX_EOF;

  __src_int_to_float_array(ibuf,p->input.data,*isamp);

  p->src_data.end_of_input=0;

  p->src_data.input_frames=*isamp/channels;
  p->src_data.data_in=p->input.data;

  p->src_data.src_ratio=p->rate/rate;

  p->src_data.data_out=p->output.data;
  p->src_data.output_frames=*osamp/channels;

  if (0!=__src_process(p->src_state,&p->src_data))
    return SOX_EOF;

  *isamp=p->src_data.input_frames_used*channels;
  *osamp=p->src_data.output_frames_gen*channels;
  __src_float_to_int_array(p->output.data,obuf,*osamp);

  return SOX_SUCCESS;
}

static int stop(sox_effect_t *effp)
{
  priv_t *p=effp->priv;

  if (NULL!=p->src_state) {
    __src_delete(p->src_state);
    p->src_state=NULL;
  }

  return SOX_SUCCESS;
}

static int kill(sox_effect_t *effp)
{
  priv_t *p=effp->priv;

  buf_cleanup(&p->output);
  buf_cleanup(&p->input);

  return SOX_SUCCESS;
}

sox_effect_handler_t const *samplerate_handler()
{
  static sox_effect_handler_t handler = {
    .name="samplerate",
    .usage=NULL,
    .flags=SOX_EFF_MCHAN,
    .getopts=getopts,
    .start=start,
    .flow=flow,
    .drain=NULL,
    .stop=stop,
    .kill=kill,
    .priv_size=sizeof(priv_t)
  };

  return &handler;
}
